# =========================
# Main file for Ruby Code Editor
#
# by Alexander Schreyer, www.alexschreyer.net, mail@alexschreyer.net
#
# DO NOT EDIT THIS FILE IN THE RUBY CODE EDITOR, IT WILL STRIP SOME OF THE CODE!!!
# =========================


require 'sketchup.rb'


## =====================


module AS_Extensions

  module AS_RubyEditor


    # Creates new class
    class RubyEditor < UI::WebDialog


        # Initialize class and callbacks
        def initialize


          ## =====================

          ## Set some variables

          # Get this file's directory w/ backcomp for "//" in file path
          @base_dir = File.dirname(__FILE__).gsub(%r{//}) { "/" }

          # Get user directory
          @user_dir = (ENV['USERPROFILE'] != nil) ? ENV['USERPROFILE'] :
            ((ENV['HOME'] != nil) ? ENV['HOME'] : @base_dir )

          # Get working directory from last file - set to user directory otherwise
          @last_file = Sketchup.read_default "as_RubyCodeEditor", "last_file"
          @last_file != nil ? @snip_dir = File.dirname(@last_file) : @snip_dir = @user_dir


          ## =====================

          ## Set up the WebDialog interface

          # Dialog parameters
          super "Ruby 代码编辑器", false, "Ruby 代码编辑器", 750, 600, 100, 100, true

          # Set HTML UI file for WebDialog
          ui_loc = File.join( @base_dir , "ui.html" )
          set_file( ui_loc )
          navigation_buttons_enabled = false
          min_width = 500
          min_height = 500


          ## =====================

          ## Show the dialog and run the new routine

          show do

            # Set version number in about dlg
            execute_script("var rceVersion = #{AS_RubyEditor::EXTVERSION.to_s}")
            execute_script("$('#version').text( rceVersion )")

            Sketchup.status_text = "#{AS_RubyEditor::EXTTITLE} | 欢迎！"

          end  # show


          ## =====================

          ## Offer a save when dialog closes

          set_on_close do

            # Some feedback
            execute_script("addResults('关闭编辑器...')")

            # Handling closing confirmation in RB rather than JS
            r = UI.messagebox "退出之前保存此文件吗？", MB_YESNO

            if r == 6 then

              execute_script("cb_save()")

            end

            Sketchup.status_text = "#{AS_RubyEditor::EXTTITLE} | 再见！"

          end  # set_on_close


          ## =====================

          ## Callback to CLEAR (NEW) editor

          add_action_callback("new") do |dlg, params|

            # Get initial code snippet - load from file
            loadcode = File.readlines(File.join( AS_RubyEditor::EXTDIR , 'as_rubyeditor' , 'templates' , 'default.rb' ))
            # Fix newlines so that they are preserved
            loadcode.each{ |i| i.gsub!(/\n/, '') }
            @initCode = loadcode.join('\n')

            # Send initial code to editor. Use only single quotes here to preserve newlines!
            script = 'editor.getDoc().setValue(\''+@initCode+'\')'
            dlg.execute_script(script)

            # Update new file names in editor
            dlg.execute_script("$('#save_name').val('无标题.rb')")
            dlg.execute_script("$('#file_name').text('无标题.rb')")
            dlg.execute_script("$('#file_name').attr('title','')")
            dlg.execute_script("$('#save_filepath').val('')")

            # Reset the editor after loading
            dlg.execute_script("editor.scrollTo(0,0)")
            dlg.execute_script("addResults('清空编辑器')")
            dlg.execute_script("editor.markClean()")
            dlg.execute_script("editor.getDoc().clearHistory()")

            # Update the MRU list (from SU prefs)
            # ... Load current MRU
            mru = []
            (1..5).each { |i|
               mru.push Sketchup.read_default("as_RubyCodeEditor", "mru#{i}", "" )
            }
            dlg.execute_script("updateMRU( '#{mru[0]}' , '#{mru[1]}' , '#{mru[2]}' , '#{mru[3]}' , '#{mru[4]}' )")

          end  # add_action_callback("new")


          ## =====================

          ## Callback to LOAD (OPEN) a file into the editor

          add_action_callback("load") do |dlg, params|

            # Get filename (dialog, parameter, default)
            case params
              when 'undefined'  # No parameter supplied
                file = UI.openpanel("打开文件", @snip_dir, "Ruby 文件|*.rb|全部文件|*.*||")
              when 'default'  # Parameter "default" opens default file
                file = File.join( AS_RubyEditor::EXTDIR , 'as_rubyeditor' , 'templates' , 'default.rb' )
              else  # File and path has been supplied by parameter
                file = params
            end

            if not File.exist?(file)
              UI.messagebox "无法加载 #{File.basename(file).to_s}， 该文件不存在（此处）。"
              return
            elsif not file
              UI.messagebox "未选择有效文件。"
              return
            end

            begin

              # Fix slashes for Windows
              file.tr!("\\","/")

              # Set file directory as current and get file details
              @snip_dir = File.dirname(file)
              Dir.chdir( @snip_dir )
              name = File.basename(file)
              extension = File.extname(file)

              # Read text from file
              f = File.new(file,"r")
              text = f.readlines.join

              # Load text into editor by encoding some parts in RB and unencoding in JS:
              # ... Encode backward slashes and single quotes in Ruby
              text.gsub!('\\', "<84JSed>")
              text.gsub!('\'', "<25SKxw>")
              text.gsub!(/\n/, "\\n")
              text.gsub!(/\r/, "\\r")
              text.gsub!(/'\'/, '\\')
              # ... Load text into variable in JS and unencode the slashes and quotes
              dlg.execute_script("tmp = '#{text}'")
              dlg.execute_script("tmp = tmp.replace(/<84JSed>/g,'\\\\')")
              dlg.execute_script("tmp = tmp.replace(/<25SKxw>/g,'\\'')")
              script = 'editor.setValue(tmp)'
              dlg.execute_script(script)

              # Reset the editor after loading
              dlg.execute_script("editor.scrollTo(0,0)")
              dlg.execute_script("addResults('文件加载: #{name}')")
              dlg.execute_script("editor.markClean()")
              dlg.execute_script("editor.getDoc().clearHistory()")

              # Update new file names in editor
              dlg.execute_script("$('#save_name').val('#{name}')")
              dlg.execute_script("$('#file_name').text('#{name}')")
              dlg.execute_script("$('#file_name').attr('title','#{file}')")
              dlg.execute_script("$('#save_filepath').val('#{file}')")

              # Update the MRU list (from SU prefs)
              # ... Load current MRU
              mru = []
              (1..5).each { |i|
                 mru.push Sketchup.read_default("as_RubyCodeEditor", "mru#{i}", "" )
              }
              # ... Don't update if file is already in there
              if not mru.include?(file)
                (5).downto(2) { |i|
                  Sketchup.write_default("as_RubyCodeEditor", "mru#{i}", mru[i-2].to_s )
                }
                Sketchup.write_default("as_RubyCodeEditor", "mru1", file )
              end
              # ... Write new MRU to editor menu
              dlg.execute_script("updateMRU( '#{mru[0]}' , '#{mru[1]}' , '#{mru[2]}' , '#{mru[3]}' , '#{mru[4]}' )")
              # ... Make a note of the last file (not necessarily mru1)
              Sketchup.write_default "as_RubyCodeEditor", "last_file", file

            rescue Exception => e

              UI.messagebox "无法打开 #{File.basename(file).to_s}。 \n\n错误: #{e}"

            end

          end  # add_action_callback("load")


          ## =====================

          ## Callback to SAVE / SAVE AS a file (and create a backup)

          add_action_callback("save") do |dlg, params|

            # Extract all parameters
            params = params.split(',')

            # Get current filename from hidden inputs
            filename = dlg.get_element_value("save_name")

            # Get filename
            if params[1] == 'true' or dlg.get_element_value("save_filepath") == ""
              file = UI.savepanel("保存文件", @snip_dir, filename)  # Allow to change save file in dialog
            else
              file = dlg.get_element_value("save_filepath")  # Get existing filename from hidden field
            end
            return if file.nil?
            file.tr!("\\","/")  # Fix slashes for Windows

            # Set file directory as current and get file details
            @snip_dir = File.dirname(file)
            Dir.chdir( @snip_dir )
            extension = File.extname(file)
            # Add RB extension if nothing is there
            file = file+".rb" if extension == ""
            name = File.basename(file)

            # Get text from editor and clean it up
            str = dlg.get_element_value("console")
            str.gsub!(/\r\n/, "\n")

            # Copy current file to backup if file already exists
            if File.exist?(file) and params[0] == 'true'
              f = File.new(file,"r")
              oldfile = f.readlines
              File.open(file+".bak", "w") { |f| f.puts oldfile }
            end

            # Write text to file
            File.open(file, "w") { |f| f.puts str }

            # Update new file names in editor
            dlg.execute_script("$('#save_name').text('#{name}')")
            dlg.execute_script("$('#file_name').text('#{name}')")
            dlg.execute_script("$('#file_name').attr('title','#{file}')")
            dlg.execute_script("$('#save_filepath').val('#{file}')")

            # Reset the editor
            dlg.execute_script("editor.markClean()")
            dlg.execute_script("addResults('保存文件: #{name}')")

            # Update the MRU list (from SU prefs)
            # ... Load current MRU
            mru = []
            (1..5).each { |i|
               mru.push Sketchup.read_default("as_RubyCodeEditor", "mru#{i}", "" )
            }
            # ... Don't update if file is already in there
            if not mru.include?(file)
              (5).downto(2) { |i|
                Sketchup.write_default("as_RubyCodeEditor", "mru#{i}", mru[i-2].to_s )
              }
              Sketchup.write_default("as_RubyCodeEditor", "mru1", file )
            end
            # ... Write new MRU to editor menu
            dlg.execute_script("updateMRU( '#{mru[0]}' , '#{mru[1]}' , '#{mru[2]}' , '#{mru[3]}' , '#{mru[4]}' )")
            # ... Make a note of the last file (not necessarily mru1)
            Sketchup.write_default "as_RubyCodeEditor", "last_file", file

          end  # add_action_callback("save")


          ## =====================

          ## Callback to EXECUTE Ruby code in SketchUp

          add_action_callback("exec") do |dlg, params|

            # Provide some status text
            dlg.execute_script( "addResults('运行代码...')" )
            Sketchup.status_text = "#{AS_RubyEditor::EXTTITLE} | 运行代码..."

            # Add some paths to loadpath variable (if those are supplied)
            lp1 = get_element_value("loadpath1")
            lp2 = get_element_value("loadpath2")
            if not lp1.empty?
              lp1 = File.expand_path(lp1)
              $LOAD_PATH << lp1 unless $LOAD_PATH.include? lp1
            end
            if not lp2.empty?
              lp2 = File.expand_path(lp2)
              $LOAD_PATH << lp2 unless $LOAD_PATH.include? lp2
            end

            # Get the code from the editor and encode it
            v = dlg.get_element_value('console').strip
            # ... Force encoding for non-UTF text (e.g. in China) w/ backcomp
            v.force_encoding('UTF-8') if v.respond_to?(:force_encoding)

            # Execute the code with eval and rescue if error
            r = nil
            rline = nil
            begin
              # ... Wrap everything in single undo if desired
              Sketchup.active_model.start_operation "RCE Code Run" if params == 'true'
              begin  # ... Evaluation under the top level binding
                r = eval( v , TOPLEVEL_BINDING )
              rescue => e
                r = e
                # could use later: rfull = e.backtrace.join('\n')
                rline = e.backtrace[0].split(":")[1].to_i + 1
                raise  # ... Pass to outer rescue clause if error
              end
            rescue  # ... If error
              Sketchup.active_model.abort_operation
              r = '运行中止。错误： ' + e
            else  # ... Commit process if no errors
              Sketchup.active_model.commit_operation if params == 'true'
            ensure  # ... Always do this
              if r === nil
                r = '无结果（未返回结果）'
              else
                r = r.to_s
                r = r + " (Line #{rline})" if rline != nil
              end
              p r  # ... Also return result to console
              # ... Format for HTML box
              r.gsub!(/ /, "&nbsp;")
              r.gsub!(/\n/, "<br>")
              r.gsub!(/'/, "&rsquo;")
              r.gsub!(/`/, "&lsquo;")
              r.gsub!(/</, "&lt;")

              # Provide some status text and return result
              dlg.execute_script("addResults('运行代码完成。 Ruby 返回： <span class=\\'hl\\'>#{r}</span>')")
              Sketchup.status_text = "#{AS_RubyEditor::EXTTITLE} | 已运行完代码"

            end

          end  # add_action_callback("exec")


          ## =====================

          ## Callback to EXECUTE (LOAD) external code (in Ruby file)

          add_action_callback("exec_load") do |dlg, params|

            # Pick an RB file
            f = UI.openpanel( "选择一个Ruby文件以从其位置运行（加载）它", @snip_dir, "Ruby 文件|*.rb|全部文件|*.*||" )

            if f

              begin

                raise "所选文件不是Ruby文件。" if File.extname(f) != ".rb"

                r = nil

                Sketchup.active_model.start_operation "外部代码运行"

                # Load this plugin and add path to search paths
                d = File.dirname( f )
                d = File.expand_path( d )
                $: << d unless $:.include? d
                load f

                Sketchup.active_model.commit_operation

                dlg.execute_script("addResults('文件运行完毕: #{f.tr!("\\","/")}')")

                # Set directory as last used and give feedback
                # Don't do this for now - messes with editor file loading

              rescue => e

                r = e
                r === nil ? r = '无结果（未返回结果）' : r = r.to_s
                # ... Format for HTML box
                r.gsub!(/ /, "&nbsp;")
                r.gsub!(/\n/, "<br>")
                r.gsub!(/'/, "&rsquo;")
                r.gsub!(/`/, "&lsquo;")
                r.gsub!(/</, "&lt;")
                dlg.execute_script("addResults('文件运行完毕： #{f.tr!("\\","/")}. Ruby 返回： <span class=\\'hl\\'>#{r}</span>')")

              end

            end

          end  # add_action_callback("exec_load")


          ## =====================

          ## Callback to CLOSE the dialog

          add_action_callback("quit") do |dlg, params|

            dlg.close

          end  # add_action_callback("quit")


          ## =====================

          ## Callback to UNDO the last grouped code execution

          add_action_callback("undo") do |dlg, params|

            Sketchup.undo

            dlg.execute_script("addResults('最后一步操作已撤消')")
            Sketchup.status_text = "#{AS_RubyEditor::EXTTITLE} | 最后一步操作已撤消"

          end  # add_action_callback("undo")


          ## =====================

          ## Callback to EXPLORE current SELECTION

          add_action_callback("sel_explore") do |dlg, params|

            sel = Sketchup.active_model.selection

            mes = ""
            mes += "已选中 "
            mes += "#{sel.length} "
            mes += sel.length == 1 ? " 个实体\n\n" : "entities"

            sel.each { |e|

              # Show useful properties
              mes += "实体： #{e.to_s}\n"
              mes += "类型： #{e.typename}\n"
              mes += "ID： #{e.entityID}\n"
              mes += "持久ID： #{e.persistent_id}\n" if Sketchup.version.to_f >= 17
              mes += "图层： #{e.layer.name}\n"
              mes += "中心位置 (x,y,z)： #{e.bounds.center}\n"
              size = e.bounds.max - e.bounds.min
              mes += "大小 (x,y,z)： #{size}\n"
              mes += "定义名称： #{e.definition.name}\n" if e.is_a? Sketchup::ComponentInstance
              mes += "父级： #{e.parent}\n"

              mes += "\n"

            }

            UI.messagebox mes , MB_MULTILINE, "列出当前选定内容的属性"

          end  # add_action_callback("sel_explore")


          ## =====================

          ## Callback to EXPLORE current selection's ATTRIBUTES

          add_action_callback("att_explore") do |dlg, params|

            sel = Sketchup.active_model.selection

            mes = ""
            mes += "已选中 "
            mes += "#{sel.length} "
            mes += sel.length == 1 ? " 个实体\n\n" : "entities"

            sel.each { |e|

              mes += "实体： #{e.to_s}\n"
              mes += "ID： #{e.entityID}\n"
              mes += "持久ID： #{e.persistent_id}\n" if Sketchup.version.to_f >= 17

              # Check for entity attributes
              if e.attribute_dictionaries
                mes += "实体属性字典：\n"
                names = ""
                e.attribute_dictionaries.each {|dic|
                  mes += "  字典名称： #{dic.name}\n"
                  dic.each { | key, value |
                    mes += "    " + key.to_s + '=' + value.to_s + "\n"
                  }
                }
              else
                mes += "未定义实体属性。\n"
              end

              # Check for component attributes
              if e.is_a? Sketchup::ComponentInstance and e.definition.attribute_dictionaries
                mes += "定义属性字典：\n"
                names = ""
                e.definition.attribute_dictionaries.each {|dic|
                  mes += "   字典名称： #{dic.name}\n"
                  dic.each { | key, value |
                    mes += "      " + key.to_s + '=' + value.to_s + "\n"
                  }
                }
              else
                mes += "没有定义属性。\n"
              end

              mes += "\n"

            }

            UI.messagebox mes , MB_MULTILINE, "列出当前选定内容的特性"

          end  # add_action_callback("att_explore")


          ## =====================

          ## Callback to INSERT the selection reference

          add_action_callback("insert_ref") do |dlg, params|

            ents = []
            Sketchup.active_model.selection.each { | e |
              # Use persistent ID when available
              Sketchup.version.to_f < 17 ? ents << e.entityID : ents << e.persistent_id
            }
            dlg.execute_script("editor.replaceSelection('#{ents}')")

          end  # add_action_callback("insert_ref")


          ## =====================

          ## Callback to show RUBY CONSOLE

          add_action_callback("show_console") do |dlg, params|

            Sketchup.send_action "showRubyPanel:"

          end  # add_action_callback("show_console")


          ## =====================

          ## Callback to show PLUGIN FOLDER

          add_action_callback("plugin_folder") do |dlg, params|

            UI.openURL("file:///#{Sketchup.find_support_file('Plugins')}")

          end  # add_action_callback("plugin_folder")


          ## =====================

          ## Callback to show HELP dialog in browser

          add_action_callback("help") do |dlg, params|

            AS_RubyEditor::browser( "#{AS_RubyEditor::EXTTITLE} - 帮助" , "https://alexschreyer.net/projects/sketchup-ruby-code-editor/" )

          end  # add_action_callback("help")


          ## =====================

          ## Callback to show REFERENCE BROWSER dialog

          add_action_callback("browser") do |dlg, params|

            AS_RubyEditor::browser( "#{AS_RubyEditor::EXTTITLE} - 参考网站" , File.join( AS_RubyEditor::EXTDIR , 'as_rubyeditor' , 'ui-browser.html' ) , true )

          end  # add_action_callback("browser")


       end # initialize


    end # class RubyEditor


    ## =====================

    ## Show local or remote website either as a WebDialog or HtmlDialog

    def self.browser( title , loc , isfile = false )

      if Sketchup.version.to_f < 17 then  # Use old WebDialog method
        @dlg = UI::WebDialog.new( title , true ,
          title.gsub(/\s+/, "_") , 1000 , 600 , 100 , 100 , true);
        @dlg.navigation_buttons_enabled = false
        isfile ? @dlg.set_file( loc ) : @dlg.set_url( loc )
        @dlg.show
      else  # Use new HtmlDialog
        @dlg = UI::HtmlDialog.new( { :dialog_title => title, :width => 1000, :height => 600,
          :style => UI::HtmlDialog::STYLE_DIALOG, :preferences_key => title.gsub(/\s+/, "_") } )
        isfile ? @dlg.set_file( loc ) : @dlg.set_url( loc )
        @dlg.show
        @dlg.center
      end

    end  # def self.browser


    ## =====================

    ## Add menu items

    unless file_loaded?(__FILE__)

      # Add main menu items
      sub = UI.menu("Window").add_submenu( "Ruby代码编辑器" )
      sub.add_item("Ruby代码编辑器") { editordlg = AS_RubyEditor::RubyEditor.new }
      sub.add_item("参考网站") { self.browser( "#{AS_RubyEditor::EXTTITLE} - 参考网站" , File.join( AS_RubyEditor::EXTDIR , 'as_rubyeditor' , 'ui-browser.html' ) , true ) }
      sub.add_item("帮助") { self.browser( "#{AS_RubyEditor::EXTTITLE} - 帮助" , "https://alexschreyer.net/projects/sketchup-ruby-code-editor/" ) }

      # Add toolbar
      as_rce_tb = UI::Toolbar.new "Ruby代码编辑器"
      as_rce_cmd = UI::Command.new("Ruby代码编辑器") { editordlg = AS_RubyEditor::RubyEditor.new }
      # One instance only version:
      # as_rce_cmd = UI::Command.new("Ruby Code Editor") { editordlg = AS_RubyEditor::RubyEditor.new unless editordlg }
      if Sketchup.version.to_i >= 16
        if RUBY_PLATFORM =~ /darwin/
          as_rce_cmd.small_icon = "img/tb_rubyeditor.pdf"
          as_rce_cmd.large_icon = "img/tb_rubyeditor.pdf"
        else
          as_rce_cmd.small_icon = "img/tb_rubyeditor.svg"
          as_rce_cmd.large_icon = "img/tb_rubyeditor.svg"
        end
      else
        as_rce_cmd.small_icon = "img/tb_rubyeditor_16.png"
        as_rce_cmd.large_icon = "img/tb_rubyeditor_24.png"
      end
      as_rce_cmd.tooltip = "Ruby代码编辑器"
      as_rce_cmd.status_bar_text = "在一个漂亮的对话框中编辑和运行Ruby脚本"
      as_rce_cmd.menu_text = "Ruby代码编辑器"
      as_rce_tb = as_rce_tb.add_item as_rce_cmd
      as_rce_tb.show

      # Tell SU that we loaded this file
      file_loaded(__FILE__)

    end  # unless


    ## =====================


  end  # module AS_RubyEditor

end  # module AS_Extensions


## =====================
